 ; Prepend this code to a text file to get an executable text file.
 ; File size limited by .COM size and the need for a line break table.

 ; First line is reserved for some HELP TEXT. No horizontal scrolling!
 ; Long lines are just clipped, so take care with 40 column modes.

 ; Use all vertical cursor and scrolling keys to navigate, ESC to exit.

 ; This version uses int 10.2 / 10.9 to be very compatible. Speed up
 ; using TURBO3 if you want.
 ; Open source freeware by Eric Auer eric@coli.uni-sb.de

 ; Almost complete rewrite or a tool that I have written in the mid-90s.
 ; Some versions of THAT tool even contained a builtin executable
 ; compressor and decompressor. For THIS version, you must UPX the .com
 ; yourself. You must even copy /b text2com.bin + thing.txt thing.txt
 ; by hand, no file creation helper contained in this version.


        org  100h

 ; -------------------------------------------------------------

start:	mov si, textbuf
	mov di, sp
	sub di, 4000	; room for 2000 line pointers - stack
	jc nomem
	cmp di, si
	jnb memokay	; enough RAM free

nomem:	mov dx, nomemmessage
	mov ah,9	; print warning message
	int 21h
	mov ax, 4c01h	; exit with errorlevel 1
	int 21h

memokay:
	mov [ds:linearray],di
	cld
	xor bx,bx
	mov cx,1900	; 2000 - stack room...
	; *** mov ax,si
	; *** stosw	; store pointer to first line
parse:	lodsb
	or al,al	; NUL
	jz didparse
	cmp al,26	; EOF
	jz didparse
	cmp al,10	; LF
	jb parse
	jz nextline
	cmp al,13	; CR
	jnz parse
nextline:	
	lodsb
	dec si
	cmp al,10
	jz skipdouble
	cmp al,13
	jz skipdouble
	jmp short storepos
skipdouble:
	inc si		; count CR+LF or LF+CR as only one linebreak
storepos:
	mov ax,si
	stosw		; store pointer to next line
	loop parse

didparse:
	; *** inc di
	; *** inc di
	mov [ds:linearraymax],di	; first entry after linearray
	sub di,[ds:linearray]
	shr di,1	; di is now max line number (minimum: 1)
	dec di
	mov [ds:maxline],di	; max line number
	xor ax,ax
	mov [ds:topline],ax	; current line number

 ; -------------------------------------------------------------

	push ds
	mov ax,40h
	mov ds,ax	; BIOS data segment

	; Interesting values in BIOS data segment:
	; 40:63 word  - 3d4 color 3b4 mono
	;             - b800 color b000 mono screen seg
	; 40:62 byte  - current page 0..7
	; 40:50 words - cursor pos for each page
	; 40:4e word  - offset current page in screen buffer
	; 40:4c word  - screen buffer page size (bytes)
	; 40:4a word  - columns on screen
	; 40:49 byte  - current video mode
	; 40:84 byte  - maximum row number (not for CGA/mono)

	mov ax,[ds:4ah]
	mov [cs:columns],ax	; columns on screen
	mov ah,0
	mov al,[ds:84h]
	inc al
	cmp al,25		; leave default 25 if "< 25"
	jb ignorerows
	mov [cs:rows],ax	; rows on screen
ignorerows:
	mov al,[ds:62h]		; current page
	mov bh,al		; page, kept from now on!
	mov bl,1eh		; color, keep from now on!
	add ax,ax
	add ax,50h		; cursor pos array offset
	mov [cs:curposp],ax	; current cursor pos pointer...

	pop ds

; Int 10 functions:
; AH = 2 BH = page DL = col DH = row - set cursor position
; AH = 9 BH = page BL = color CX = count AL = character - display character

clearscreen:
	xor dx,dx		; DL=column DH=row
clsloop:
	mov ah,2		; move cursor
	int 10h			; "home" the cursor
	mov ax,0920h		; print space char
	mov cx,[cs:columns]	; print a whole row
				; page/color BH/BL still set
	int 10h			; print spaces to clear screen
	inc dh			; next line
	cmp dh,[cs:rows]	; end reached?
	jb clsloop

 ; -------------------------------------------------------------

	mov si,[cs:topline]	; current first visible line
	jmp nesc		; start by painting the screen!

mainloop:
	mov si,[cs:topline]	; current first visible line
        mov ah,0        	; read key
        int 16h         	; keyboard BIOS
				; now AH is scancode and AL is ASCII
nhelp:  cmp ah,48h      ; UP
        jnz nup
        or si,si
        jz nup		; only go up if not at top already
        dec si
nup:    cmp ah,50h      ; DN
        jnz ndn
        cmp si,[cs:maxline]
        jnb ndn		; only go down if not end reached
        inc si
ndn:    cmp ah,49h      ; PGUP
        jnz npgup
	inc si		; one less (status line)
        cmp si,[cs:rows]
        jb upmost	; avoid underflow when going 1 page up
        sub si,[cs:rows]
npgup:  cmp ah,51h      ; PGDN
        jnz npgdn
	dec si		; one less (status line)
        add si,[cs:rows]
        cmp si,[cs:maxline]
        jae dnmost	; avoid overflow when going 1 page down
npgdn:  cmp ah,47h      ; POS1
        jnz npos1
upmost: xor si,si	; uppermost row
npos1:  cmp ah,4fh      ; END
        jnz nende
dnmost: mov si,[cs:maxline]
nende:  cmp ah,01h      ; ESC (ASCII 27) means exit
        jnz nesc

	mov dl,0		; left column
	mov dh,[cs:rows]
	dec dh			; go to last row
	mov ah,2		; move cursor
	int 10h			; "home" the cursor
	mov ax,0920h		; print space char
	mov cx,[cs:columns]	; print a whole row
				; page/color BH/BL still set
	int 10h			; print spaces to clear line
        mov ax,4c00h	; leave program!
        int 21h

nesc:
	mov ax,[cs:maxline]
	cmp ax,[cs:rows]
	jb noscroll	; no scrolling at all if text is max 1 screen
	sub ax,[cs:rows]
	inc ax
	inc ax
	cmp si,ax	; would we see several empty lines after EOF ?
	jb scrollthere	; else everything is okay
	mov si,ax	; tune position to hide part after EOF
	jmp short scrollthere
noscroll:
	mov si,0
scrollthere:
	mov [cs:topline],si	; update value

 ; -------------------------------------------------------------

	add si,si
	add si,[cs:linearray]	; translate to pointer into linearray
	xor dx,dx		; cursor home position
	mov ah,2		; position cursor
	int 10h			; BH and DX are already set
	mov ax,textbuf		; help line is always first
	call printline		; *** from AX on...
	inc dh			; next line
showloop:
	mov ah,2		; position cursor
	int 10h			; BH and DX are already set
	lodsw			; fetch a line
	cmp si,[cs:linearraymax]	; no more lines available?
	jae emptyline
	call printline		; *** from AX, until linebreak / border
doneline:
	inc dh			; next line
	cmp dh,[cs:rows]
	jb showloop
	jmp mainloop		; done? ask for new user input

emptyline:
	mov cx,[cs:columns]	; how many
	mov ax,0920h		; print spaces
	int 10h			; fill, BH already set.
	mov cx,1
	mov ax,097eh		; tilde marks empty line
	int 10h			; print char, BH already set.
	jmp short doneline
	
 ; Int 10 functions:
 ; AH = 2 BH = page DL = col DH = row - set cursor position
 ; AH = 9 BH = page BL = color CX = count AL = character - display character

 ; -------------------------------------------------------------

 ; should process BS (8), TAB (9), LF (10), CR (13) specially here.
 ; simplified a lot...

printline:	; print line from [ax], preserve SI, CX, DX, use BH
	push si
	push cx
	push dx
	mov si,ax		; source pointer
	mov dl,0		; column counter
	mov cx,[cs:columns]	; how many
	mov ax,0920h		; print spaces (clear before print)
	int 10h			; fill, BH already set.
prnloop:
	lodsb			; get char
	cmp al,13		; CR
	jz prndone
	cmp al,10		; LF
	jz prndone
	cmp al,9		; TAB
	jz prntab
	mov cx,1		; print char once
	mov ah,9		; print char
	int 10h			; BX still as give
	inc dl			; next column
prnwrap:
	cmp dl,[cs:columns]	; cut line
	jz prndone
	mov ah,2		; move cursor to the right...
	int 10h			; needed!
	jmp short prnloop

prntab: mov ah,2		; move cursor
	int 10h
	mov cx,1		; print one...
	mov ax,0920h		; ...space
	int 10h
	inc dl
	test dl,7
	jz prnwrap		; reached a tab stop
	jmp short prntab	; otherwise, add more spaces

prndone:
	pop dx
	pop cx
	pop si
	ret

 ; -------------------------------------------------------------

rows:		dw 25
columns:	dw 80
curposp		dw 50h		; pointer to where cursor pos is

linearray:	dw 0e000h	; where line pointer array is
linearraymax:	dw 0efffh	; pointer to after line pointer array
maxline:	dw 10		; maximum line number
topline:	dw 0		; which line is on top now?

nomemmessage:	db "Not enough RAM to load, sorry. 4k stack needed."

textbuf:
	db "Exit: ESC. Scroll: (Page)Up/Down, Home/End. "
	db "Freeware by: eric@coli.uni-sb.de",13,10
